/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Thu Jan/02/2003 11:47 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <Option.h>
#include "playlist.h"

using namespace std;
using namespace rs;
using namespace rs::cmdline;
using namespace rs::err;
using namespace rs::str;

static ProgramInfo             fixPlaylist_Info("fixPlaylist", 1, 15, "Fixes paths in playlists (M3U and PLS format)");
static CmdLineArgument         validList("validList", "playlist containing ALL your songs with valid paths");
static RepeatedCmdLineArgument invalidList("invalidList", "playlist containing SOME songs with invalid paths", acceptAnyArgument);

static CmdLineOption         autoSelect('a', "automatically select changed path, if filename is equal (path may differ)");
static CmdLineArgumentOption equalPaths('e', "dirs", "auto-select only if the last <dirs> subdirectories are equal");
static CmdLineOption         delDupes('d', "automatically delete duplicate entries");

static CmdLineArgumentOption askAbove('s', "value", "ask only if match-value >= <value>", "0.1");

static CmdLineOption force_PLS('P', "convert to PLS format");
static CmdLineOption force_M3U('M', "convert to M3U format");

static CmdLineOption keepExtInfo('k', strf("keep extended playlist info in %s", invalidList.get_name().c_str()).c_str());
static CmdLineOption copyExtInfo('c', strf("always copy extended playlist info from %s", validList.get_name().c_str()).c_str());

static CmdLineArgumentOption missingList('m', "missingList", "write unreplaced entries to <missingList>");
static CmdLineOption appendMissing('A', "append to <missingList> [Default is to overwrite]");

static CmdLineOption         findDupes('D', "Dupe-Finder (writes possible Dupes into <invalidList>)");
static CmdLineArgumentOption minDupeMatchValue('v', "value", "minimal dupe name-match percentage", "0.8");
static CmdLineArgumentOption maxTimeDiff('t', "seconds", "maximum time difference for dupes", "5");

//  -----------------------------------------
//      int main(int argc, char *argv[])
//  -----------------------------------------
int main(int argc, char *argv[]) {
    try {
        fixPlaylist_Info.init(argc, argv);
        fixPlaylist_Info.neededBy(autoSelect, equalPaths);

        {
            CmdLineOption *Options_FixPlaylist[] = { &autoSelect, &equalPaths, &delDupes, &askAbove, 0 };
            CmdLineOption *Options_DupeFinder[] = { &findDupes, &minDupeMatchValue, &maxTimeDiff, 0 };

            fixPlaylist_Info.exclude(Options_FixPlaylist, Options_DupeFinder);
        }

        PlayList valid(validList.get());
        valid.load(cout);

        if (findDupes) {
            PlayList possibleDupes(invalidList.get());

            valid.addDupesTo(possibleDupes, minDupeMatchValue.getDouble(), maxTimeDiff.getNumeric());
            possibleDupes.save(cout);
        }
        else {
            PlayList       missing(missingList ? missingList.get() : "dummylist");
            list<PlayList> invalids;

            if (appendMissing && missingList) {
                missing.load(cout);
            }

            while (invalidList) invalids.push_back(PlayList(invalidList.get()));
            if (invalids.empty()) throw Error("Nothing to do.");

            double ask_above = askAbove.getDouble();
            for (list<PlayList>::iterator i=invalids.begin(); i!=invalids.end(); ++i) {
                try {
                    i->load(cout);
                    i->correct(valid, missing, cout, delDupes, autoSelect, equalPaths.getNumeric(), copyExtInfo, keepExtInfo, ask_above);
                    if (i->is_pls_format()) {
                        if (force_M3U) i->set_pls_format(false);
                    }
                    else {
                        if (force_PLS) i->set_pls_format(true);
                    }

                    i->save(cout);
                    cout << endl;
                }
                catch (Error& err) {
                    err.print(cerr);
                    cerr << "Ignored. Continuing.\n";
                }
            }

            if (missingList) {
                if (force_PLS) missing.set_pls_format(true);
                missing.save(cout); // save playlist with missing songs
            }
        }

        return EXIT_SUCCESS;
    }
	catch (Error& err) {
		err.print(cerr);
	}
	catch (...) {
		Error err("Unknown exception");
		err.print(cerr);
	}

    return EXIT_FAILURE;
}

