//  ======================================================================== //
//                                                                           //
//    File      : match.cpp                                                  //
//    Purpose   :                                                            //
//    Time-stamp: <Sat May/04/2002 11:53 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) January 2002 by Ralf Westram                                       //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#include <rs_assert.h>
#include <Tools.h>

#include "match.h"
#include "quickSearch.h"


using namespace std;
using namespace rs;
using namespace rs::str;

namespace rs {

    //  ----------------------------------------------------------------------------------
    //      double calcMatchValue_old(const char *s1, int l1, const char *s2, int l2)
    //  ----------------------------------------------------------------------------------
    double calcMatchValue_old(const char *s1, int l1, const char *s2, int l2) {
        int val = 0;

        for (int p1=0; p1<l1; ++p1) {
            int longestMatch = 0;

            for (int p2=0; p2<l2; ++p2) {
                if (s1[p1]==s2[p2]) { // equal char found
                    int match;

                    for (match         = 1;
                         s1[p1+match] == s2[p2+match] && s1[p1+match]!=0;
                         ++match) ;

                    longestMatch = gen_algo::max(longestMatch, match);
                }
            }
            val += longestMatch;
        }

        double Val = double(val)/(l1*l2);

//         if (val>1000 || (stristr(s2, "thiev") && stristr(s2, "glass"))) {
//             cout << strf("compared with '%s'\n"
//                          "val=%i Val=%f\n",
//                          s2, val, Val);
//         }

        //         return double(val)/(l1*l2);
        return Val;
        //     return double(val);
    }


    //  ----------------------------------------------------------------------
    //      double calcMatchValue(const QuickSearchString& qss, string s)
    //  ----------------------------------------------------------------------
    double calcMatchValue(const QuickSearchString& qss, string s) {
        static char   *buffer1     = 0;
        static size_t  buffersize1 = 0;

        size_t len1 = qss.getString().length();

        if (buffersize1 <= len1) {
            if (buffer1) delete [] buffer1;
            buffersize1 = len1+50;
            buffer1     = new char[buffersize1];
        }
        strcpy(buffer1, qss.getString().c_str());

        static char   *buffer2     = 0;
        static size_t  buffersize2 = 0;

        size_t len2 = s.length();

        if (buffersize2 <= len2) {
            if (buffer2) delete [] buffer2;
            buffersize2 = len2+50;
            buffer2     = new char[buffersize2];
        }
        strcpy(buffer2, s.c_str());

        double result = 0;

        while (1) {
            size_t max_match       = 0;
            size_t max_match_chars = 0;
            size_t max_match_pos1  = 0;
            size_t max_match_pos2  = 0;

            for (size_t pos2 = 0; buffer2[pos2]; ++pos2) {
                int ch = buffer2[pos2];
                if (QuickSearchString::is_indexed(ch)) {
                    CharOccurances::const_iterator oe = qss.end(ch);
                    for (CharOccurances::const_iterator o = qss.begin(ch); o != oe; ++o) {
                        size_t pos1  = *o;
                        size_t off   = 1;
                        size_t chars = isNonWhite(ch);
                        while (buffer1[pos1+off] && buffer1[pos1+off] == buffer2[pos2+off]) {
                            chars += isNonWhite(buffer1[pos1+off]);
                            ++off;
                        }
                        if (chars>max_match_chars) {
                            max_match       = off;
                            max_match_chars = chars;
                            max_match_pos1  = pos1;
                            max_match_pos2  = pos2;
                        }
                    }
                }
                else {
                    for (size_t pos1 = 0; pos1<len1; ++pos1) {
                        if (buffer1[pos1] == ch) {
                            size_t off   = 1;
                            size_t chars = isNonWhite(ch);
                            while (buffer1[pos1+off] && buffer1[pos1+off] == buffer2[pos2+off]) {
                                chars += isNonWhite(buffer1[pos1+off]);
                                ++off;
                            }
                            if (chars>max_match_chars) {
                                max_match       = off;
                                max_match_chars = chars;
                                max_match_pos1  = pos1;
                                max_match_pos2  = pos2;
                            }
                        }
                    }
                }
            }

            // #define SHOW_NEWMATCH_VALUES
#if defined(SHOW_NEWMATCH_VALUES)
            cout << "qss='" << qss.getString() << "'\n";
            cout << "s  ='" << s << "'\n";
            cout << "max_match=" << max_match << "\n";
            cout << "max_match_chars=" << max_match_chars << "\n";
            cout << "max_match_pos1=" << max_match_pos1 << " '" << string(buffer1+max_match_pos1, max_match) << "'\n";
            cout << "max_match_pos2=" << max_match_pos2 << " '" << string(buffer2+max_match_pos2, max_match) << "'\n";
#endif // SHOW_NEWMATCH_VALUES

            if (max_match_chars < 3) break;

            for (size_t p = max_match_pos1; p<(max_match_pos1+max_match); ++p) buffer1[p] = 0;
            strcpy(buffer2+max_match_pos2, buffer2+max_match_pos2+max_match);

            result += 2*max_match_chars;
        }

        size_t overallChars = countNonWhite(qss.getString())+countNonWhite(s);

#if defined(SHOW_NEWMATCH_VALUES)
        cout << "result=" << result << " overallChars=" << overallChars << "\n";
#endif // SHOW_NEWMATCH_VALUES

        result /= overallChars;

#if defined(SHOW_NEWMATCH_VALUES)
        cout << "result=" << result << "\n";
#endif // SHOW_NEWMATCH_VALUES

        return result;
    }

    //  ----------------------------------------------------------------------------------------------------------
    //      inline bool equal(const string& s1, size_t p1, size_t n1, const string& s2, size_t p2, size_t n2)
    //  ----------------------------------------------------------------------------------------------------------
    inline bool equal(const string& s1, size_t p1, size_t n1, const string& s2, size_t p2, size_t n2) {
        if (n1 == string::npos) n1 = s1.length()-p1;
        if (n2 == string::npos) n2 = s2.length()-p2;

        if (n1 != n2) return false;
        return strnicmp(s1.data()+p1, s2.data()+p2, n1) == 0;
    }

    // ------------------------------------------------------------------------------------
    //      bool filenames_equal(const string& p1, const string& p2, int& equal_Paths)
    // ------------------------------------------------------------------------------------
    // returns true, if filenames (not paths) of p1 and p2 are equal
    // equal_Paths is set to the number of equal subdirectories (starting from end)
    bool filenames_equal(const string& p1, const string& p2, int& equal_Paths) {
        const char *sep    = "\\/";
        size_t      slash1 = p1.find_last_of(sep);
        size_t      slash2 = p2.find_last_of(sep);

        if (slash1 != string::npos && slash2 != string::npos) {
            if (equal(p1, slash1+1, string::npos, p2, slash2+1, string::npos)) {
                equal_Paths = 0;
                // count equal paths :
                while (1) {
                    size_t fslash1 = p1.find_last_of(sep, slash1-1);
                    size_t fslash2 = p2.find_last_of(sep, slash2-1);

                    if (fslash1 == string::npos || fslash2 == string::npos) break;
                    if (!equal(p1, fslash1+1, slash1-fslash1-1, p2, fslash2+1, slash2-fslash2-1)) break; // subdir differs

                    ++equal_Paths;
                    slash1 = fslash1;
                    slash2 = fslash2;
                }
                return true;
            }
        }
        return false;
    }
};





