//  ======================================================================== //
//                                                                           //
//    File      : playlist.cpp                                               //
//    Purpose   :                                                            //
//    Time-stamp: <Thu Jul/25/2002 22:14 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) January 2002 by Ralf Westram                                       //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#include <fstream>

#include <rs_assert.h>

#include "playlist.h"
#include "match.h"
#include "quickSearch.h"

using namespace std;
using namespace rs;
using namespace rs::str;
using namespace rs::err;
using namespace rs::file;

#define TAG  "#EXTM3U"
#define TAG2 "#EXTINF:"

#define TAGLEN  7
#define TAG2LEN 8

// start of implementation of class PlaylistEntry:

// ------------------------------------------------------------------------------------
//      void PlaylistEntry::write(ostream& out, bool pls_format, int index) const
// ------------------------------------------------------------------------------------
void PlaylistEntry::write(ostream& out, bool pls_format, int index) const  {
    if (!deleted) {
        if (pls_format) {
            out << "file" << index << '=' << path << '\n';
            if (extinf) {
                string extinf2 = *extinf;
                if (strncmp(extinf2, TAG2, TAG2LEN) == 0) {
                    extinf2.erase(0, TAG2LEN);
                }

                size_t komma = extinf2.find_first_of(',');
                if (komma) {
                    out << "title" << index << "=" << extinf2.substr(komma+1) << '\n';
                    out << "length" << index << "=" << extinf2.substr(0, komma) << '\n';
                }
                else {
                    out << "title" << index << "=" << extinf2 << '\n';
                }
            }
        }
        else {
            if (extinf) out << *extinf << '\n';
            out << path << '\n';
        }
    }

}

// -end- of implementation of class PlaylistEntry.


// -------------------------------------------------------------------------------
//      inline string buildExtInfo(const string& length, const string& title)
// -------------------------------------------------------------------------------
inline string buildExtInfo(const string& length, const string& title) {
    string ext_info = "";

    if (title.length()) {
        int len                  = 0;
        if (length.length()) len = atoi(length.c_str());
        ext_info                 = string(TAG2)+strf("%i", len)+','+title;
    }

    return ext_info;
}

// --------------------------------------------------
//      void PlayList::load(ostream& screen_out)
// --------------------------------------------------
void PlayList::load(ostream& screen_out) {
    InputFile in(filename);
    string    line;
    int       entries_expected = -1; // no info
    bool      gotLine          = false;

    screen_out << "Reading playlist '" << filename << "'\n";
    {
        bool header_scanned = false;

        gotLine = in.getLine(line);
        if (gotLine) {
            if (line[0]=='#') {
                if (line!=TAG) in.throw_Error(TAG " awaited.");
                screen_out << "Detected extended m3u format.\n";
                header_scanned = true;
                gotLine        = in.getLine(line);
            }
            else if (line == "[playlist]") {
                pls_format = true;
                screen_out << "Detected pls format.\n";

                gotLine = in.getLine(line);
                if (gotLine) {
                    if (strnicmp(line, "numberofentries=", 16) == 0) {
                        entries_expected = atoi(line.c_str()+16);
                        header_scanned   = true;
                        gotLine          = in.getLine(line);
                    }
                    else {
                        header_scanned = true;
                    }
                }
            }
            else {
                screen_out << "Assuming m3u format.\n";
                header_scanned = true; // playlist w/o header
            }
        }
        else {
            screen_out << "Assuming input is an empty playlist in m3u format.\n";
            header_scanned = true;
        }

        if (!header_scanned) in.throw_Error("Unexpected end of file behind header");
    }

    if (pls_format) {
        int    last_seen_index = 0;
        string mp3             = "";
        string title           = "";
        string length          = "";

        while (gotLine) {
            cropString(line);

            if (line.length()) {
                size_t off = 0;
                bool   ok  = false;

                if (strnicmp(line, "file", 4)        == 0) off = 4;
                else if (strnicmp(line, "title", 5)  == 0) off = 5;
                else if (strnicmp(line, "length", 6) == 0) off = 6;

                if (off) {
                    size_t Equal = line.find_first_of('=', off);
                    if (Equal != string::npos) {
                        int index = atoi(line.c_str()+off);
                        if (index>0) {
                            if (last_seen_index != index) {
                                if (mp3.length()) {
                                    string ext_info = buildExtInfo(length, title);
                                    entries.push_back(PlaylistEntry(mp3, ext_info.length() ? &ext_info : 0));
                                }
                                last_seen_index = index;
                                mp3 = title = length = "";
                            }

                            string rest = line.substr(Equal+1);
                            switch (off) {
                                case 4: mp3    = rest; break;
                                case 5: title  = rest; break;
                                case 6: length = rest; break;
                                default : assert(0); break;
                            }
                            ok = true;
                        }
                    }
                }

                if (!ok && line.length() && line[0]) cerr << "Unexpected content ignored: '" << line << "'\n";
            }
            gotLine = in.getLine(line);
        }
        // write last entry
        if (mp3.length()) {
            string ext_info = buildExtInfo(length, title);
            entries.push_back(PlaylistEntry(mp3, ext_info.length() ? &ext_info : 0));
        }
    }
    else { // m3u format
        string ext_info = "";

        while (gotLine) {
            cropString(line);
            if (line[0] == '#') {
                if (ext_info.length()) screen_out << "Duplicated extended playlist info skipped\n";
                ext_info = line;
            }
            else {
                if (line.length() > 0) {
                    entries.push_back(PlaylistEntry(line, ext_info.length() ? &ext_info : 0));
                    ext_info = "";
                }
            }
            gotLine = in.getLine(line);
        }
    }
}

// --------------------------------------------------------
//      void PlayList::save(ostream& screen_out) const
// --------------------------------------------------------
void PlayList::save(ostream& screen_out) const {
    {
        ofstream out(filename.c_str());
        screen_out << "Saving '" << filename << "'\n";

        if (!out.good()) throw IOError(filename);
        if (pls_format) {
            out << "[playlist]\n"
                << "numberofentries=" << entries.size() << "\n";
        }
        else {
            if (hasExtInfo()) out << TAG "\n";
        }

        int index = 1;
        for (list<PlaylistEntry>::const_iterator e=entries.begin(); e != entries.end(); ++e, ++index) {
            e->write(out, pls_format, index);
        }
        if (pls_format) out << char(0) << '\n';
    }
    size_t extension_start = filename.find_last_of('.');
    string rename_to;
    if (extension_start == string::npos) {
        rename_to = filename+correctExtension();
    }
    else {
        rename_to = filename.substr(0, extension_start)+correctExtension();
    }

    if (rename_to != filename) {
        if (rename(filename.c_str(), rename_to.c_str()) == -1) {
            throw IOError(filename+" -> "+rename_to, "rename");
        }
        screen_out << "Playlist had wrong extension. Renamed to '" << rename_to << "'\n";
    }
}

#undef TAG

// --------------------------------------------------------------------------------------------------------------------
//      void PlayList::getBestMatchingEntries(const PlaylistEntry& entry, BestEntries& best, unsigned count) const
// --------------------------------------------------------------------------------------------------------------------
void PlayList::getBestMatchingEntries(const PlaylistEntry& entry, BestEntries& best, unsigned count) const {
    // best entries are stored at end of 'best'
    string  path1 = entry.Path();
    string *ext1  = entry.Extinf() ? new string(*entry.Extinf()) : 0;

    strlwr(path1);
    if (ext1) strlwr(*ext1);

    for (list<PlaylistEntry>::const_iterator e=entries.begin(); e != entries.end(); ++e) {

        string  path2 = e->Path();
        string *ext2  = e->Extinf() ? new string(*e->Extinf()) : 0;

        strlwr(path2);
        if (ext2) strlwr(*ext2);

        double val             = calcMatchValue_old(path1.c_str(), path1.length(), path2.c_str(), path2.length());
        if (ext1 && ext2) val += calcMatchValue_old(ext1->c_str(), ext1->length(), ext2->c_str(), ext2->length());

//         if (stristr(path2.c_str(), "thiev") && stristr(path2.c_str(), "glass")) {
//             cout << "val(gesamt)=" << val << "\n";
//         }

        if (best.size()<count || best.begin()->first<val) {
            best.insert(pair<double, PlaylistEntry>(val, *e));
            //best[val] = *e;
            if (best.size()>count) best.erase(best.begin()->first);
         }

        delete ext2;
    }

    delete ext1;
}

// -------------------------------------------------------------
//      static string skipPathAndExtension(const string& s)
// -------------------------------------------------------------
static string skipPathAndExtension(const string& s) {
    size_t nameStart                         = s.find_last_of('\\');
    if (nameStart == string::npos) nameStart = 0;
    else nameStart++;

    size_t dot = s.find_last_of('.');
    size_t len = (dot == string::npos || dot <= nameStart) ? string::npos : (dot-nameStart);

    return string(s, nameStart, len);
}
// ----------------------------------------------------
//      static string skipextinfo(const string& s)
// ----------------------------------------------------
// static string skipextinfo(const string& s) {
//     if (strnicmp(s, TAG2, TAG2LEN) == 0) {
//         size_t komma = s.find(',', 8);
//         if (komma != string::npos) {
//             return string(s, komma, s.length()-komma);
//         }
//     }
//     return s;
// }

// ------------------------------------------------------------
//      static int getDuration(const PlaylistEntry& entry)
// ------------------------------------------------------------
static int getDuration(const PlaylistEntry& entry) {
    const string *ext = entry.Extinf();
    if (ext) {
        size_t colon = ext->find(':');
        if (colon == 7) {
            size_t komma = ext->find(',', colon);
            if (komma != string::npos) {
                int t = atoi(ext->c_str()+colon+1);
                return t;
            }
        }
    }
    return 0;
}

//         string            used_ext1 = ext1 ? skipextinfo(*ext1) : "";
        //         QuickSearchString qs_ext1(used_ext1);
            //             string *ext2  = e2->Extinf() ? new string(*e2->Extinf()) : 0;


            //                 double val = calcMatchValue(name1.c_str(), name1.length(), name2.c_str(), name2.length());
            //                 if (ext1 && ext2) {
            //                     double val2       = calcMatchValue(ext1->c_str(), ext1->length(), ext2->c_str(), ext2->length());
            //                     if (val2>val) val = val2;
            //                 }
            // use extended info for matching :
            //
            //             if (ext1 && ext2) {
            //                 string used_ext2  = skipextinfo(*ext2);
            //                 strlwr(used_ext2);
            //                 double val2       = calcMatchValue(qs_ext1, used_ext2);
            //                 if (val2>val) val = val2;
            //             }

// -------------------------------------------------------------------------------------------------
//      void PlayList::addDupesTo(PlayList& dupes, double minMatchValue, int maxTimeDifference)
// -------------------------------------------------------------------------------------------------
void PlayList::addDupesTo(PlayList& dupes, double minMatchValue, int maxTimeDifference) {
    bool firstAtAll = true;

    for (list<PlaylistEntry>::const_iterator e1=entries.begin(); e1 != entries.end(); ++e1) {
        bool   first = true;
        string path1 = e1->Path(); strlwr(path1);
        string name1 = skipPathAndExtension(path1);

#if defined(SHOW_NEWMATCH_VALUES)
        cout << "name1='" << name1 << "'\n";
        if (name1.find("at night") == string::npos) continue;
#endif // SHOW_NEWMATCH_VALUES

        QuickSearchString qs_name1(name1);
        int               duration1 = getDuration(*e1);

        list<PlaylistEntry>::const_iterator e2 = e1;
        advance(e2, 1);

        for (; e2 != entries.end(); ++e2) {
            int   duration2 = getDuration(*e2);

            if (duration1 && duration2) { // duration defined for both songs
                int difference = abs(duration1-duration2);
                if (difference>maxTimeDifference && // difference to big
                    difference<(10*60*60)) // but shorter that 10 hours (to detect illegal times)
                {
                    continue; // test next entry
                }
            }

            string path2 = e2->Path();
            strlwr(path2);
            string name2 = nonnull(strrchr(path2.c_str(), '\\'))+1;
            double val   = calcMatchValue(qs_name1, name2);

            if (val >= minMatchValue) {
                if (first) {
                    if (firstAtAll) {
                        firstAtAll = false;
                    }
                    else {
                        dupes.addSeparator();
                        cout << "--------------\n";
                    }

                    first = false;
                    dupes.add(*e1);
                    cout << path1 << "\n";
                }
                dupes.add(*e2);
                cout << val << ' ' << path2 << "\n";
            }
        }

#if defined(SHOW_NEWMATCH_VALUES)
        exit(1);
#endif // SHOW_NEWMATCH_VALUES
    }
}

// ------------------------------------------------
//      int PlayList::deleteDuplicateEntries()
// ------------------------------------------------
int PlayList::deleteDuplicateEntries() {
    int deleted = 0;
    for (list<PlaylistEntry>::iterator e=entries.begin(); e != entries.end(); ++e) {
        for (list<PlaylistEntry>::iterator f=e; f != entries.end(); ++f) {
            if (f != e && f->Path() == e->Path() && !f->is_del() && !e->is_del()) {
                if (e->Extinf()) {
                    ++deleted;
                    f->del();
                }
                else {
                    ++deleted;
                    e->del();
                }
            }
        }
    }
    return deleted;
}

// --------------------------------------------------------------------------------
//     static string ask_User(const string& question)
// --------------------------------------------------------------------------------
string ask_User(const string& question) {
    cerr << question << endl;

    string answer;
    getline(cin, answer, '\n');
    return answer;
}

static bool   askAboveChanged = false;
static double userAskAbove;

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//      void PlayList::correct(const PlayList& valid, PlayList& missing, ostream& screen_out, bool delDupes, bool autoSelect, int equalPaths, bool copyInfo, bool keepInfo, double askAbove)
// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void PlayList::correct(const PlayList& valid, PlayList& missing, ostream& screen_out, bool delDupes, bool autoSelect, int equalPaths, bool copyInfo, bool keepInfo, double askAbove) {
    int corrected_paths    = 0;
    int corrected_infos    = 0;
    int deleted_paths      = 0;
    int dups_deleted_paths = 0;

    if (askAboveChanged) {
        askAbove = userAskAbove;
        cerr << "Minimum match value set to " << askAbove << "\n";
    }
    else {
        cerr << "Using command line minimum match value (" << askAbove << ")\n";
    }

    if (delDupes) dups_deleted_paths += deleteDuplicateEntries();

    for (list<PlaylistEntry>::iterator e=entries.begin(); e != entries.end(); ++e) {
        if (!e->is_del()) {
            BestEntries best;
            valid.getBestMatchingEntries(*e, best, 50*9 /*5*9*/);

            const PlaylistEntry *wanted = 0;
            if (best.rbegin()->second.Path() == e->Path()) {
                if (copyInfo) {
                    wanted = &best.rbegin()->second;
                }
            }
            else {
                if (autoSelect) { // try auto-select
                    int equalPaths_max    = 0;
                    int equalPaths_needed = equalPaths;

                    for (BestEntries::reverse_iterator m = best.rbegin(); m != best.rend(); ++m) {
                        int equal_Paths = 0;
                        if (filenames_equal(e->Path(), m->second.Path(), equal_Paths)) {
                            if (equal_Paths>equalPaths_max && equal_Paths>equalPaths_needed) {
                                equalPaths_max = equal_Paths;
                                wanted         = &m->second;
                            }
                        }
                    }

                    if (wanted) {
                        cerr << "Song: '" << e->Path() << "'\n";
                        cerr << "Auto: '" << wanted->Path() << "'\n";
                    }
                }

                bool skip = false;
                if (!wanted) {
                    BestEntries::const_reverse_iterator be = best.rbegin();
                    if (be->first < askAbove) {
                        cerr << "Skip: '" << e->Path() << "' (best=" << be->first << ")\n";
                        skip = true;
                    }
                }

                if (!wanted && !skip) {
                    int  first_idx = 0;
                    bool chosen    = false;

                    while (!chosen) {
                        {
                            BestEntries::reverse_iterator m = best.rbegin();
                            advance(m, first_idx);

                            if (m == best.rend()) {
                                if (first_idx == 0)  {
                                    cerr << "No replacements found - skipping entry!\n";
                                    break;
                                }
                                cerr << "Restarting from best matching replacement:\n";
                                first_idx = 0;
                                continue;
                            }

                            int displayed = 0;
                            if (first_idx == 0) {
                                cerr << "-------------------------------------------------------------------------------\n";
                                cerr << "Song: '" << e->Path() << "'\n";
                                cerr << "-------------------------------------------------------------------------------\n";
                            }
                            else {
                                cerr << "-------------------------------------------\n";
                                cerr << "More replacements:\n";
                                cerr << "-------------------------------------------\n";
                            }
                            for (; displayed<9 && m != best.rend(); ++m, ++displayed) {
                                cerr << (displayed+1) << ": " << m->second.Path() << " (" << m->first << ")\n";
                            }
                        }

                        bool answered  = false;
                        while (!answered) {
                            string answer = ask_User("Enter number to replace with (ENTER=skip;'d'=delete;'m'=show more;'l'=set minmatch):");
                            int    num    = atoi(answer.c_str());
                            if (num>0) {
                                BestEntries::reverse_iterator select = best.rbegin();
                                advance(select, first_idx+num-1);

                                if (select != best.rend()) wanted = &select->second;
                                if (wanted) chosen = answered = true;
                            }
                            else if (answer == "d") {
                                e->del();
                                chosen = answered = true;
                                ++deleted_paths;
                            }
                            else if (answer == "m") {
                                answered  = true;
                            }
                            else if (answer == "l") {
                                string answer2 = ask_User("Enter minimum match value");
                                if (answer2 == "") {
                                    userAskAbove = 0.0;
                                }
                                else {
                                    userAskAbove = atof(answer2.c_str());
                                }
                                askAbove        = userAskAbove;
                                askAboveChanged = true;
                                cerr << "Minimum match value set to " << askAbove << "\n";

                                answered = true;
                            }
                            else if (answer == "") {
                                chosen = answered = true;
                            }

                            if (!answered) cerr << "Illegal answer '" << answer << "'\n";
                        }

                        first_idx += 9;
                    }

                    if (!wanted) missing.add(*e); // store in "missing"-playlist
                }
            }

            if (wanted) {
                if (e->Path() != wanted->Path()) {
                    e->set_Path(wanted->Path());
                    ++corrected_paths;
                }

                if (!e->Extinf() || !keepInfo) {
                    if (wanted->Extinf() && (!e->Extinf() || (*e->Extinf() != *wanted->Extinf()))) {
                        e->set_Extinf(wanted->Extinf());
                        ++corrected_infos;
                    }
                }
            }
        }
    }
    if (delDupes) dups_deleted_paths += deleteDuplicateEntries();

    screen_out << "Corrected entries = " << corrected_paths;
    screen_out << " - Corrected infos = " << corrected_infos << "\n";
    screen_out << "Deleted entries = " << deleted_paths;
    screen_out << " - Deleted dupes = " << dups_deleted_paths << "\n";
}





