//  ======================================================================== //
//                                                                           //
//    File      : playlist.h                                                 //
//    Purpose   :                                                            //
//    Time-stamp: <Wed Nov/13/2002 08:43 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) January 2002 by Ralf Westram                                       //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef PLAYLIST_H
#define PLAYLIST_H

#ifndef LIST
#include <list>
#endif
#ifndef __MAP__
#include <map>
#endif

#ifndef INPUTFILE_H
#include <InputFile.h>
#endif


namespace rs {

    //  ----------------------------
    //      class PlaylistEntry
    //  ----------------------------
    class PlaylistEntry {
    private:
        std::string *extinf;
        std::string  path;
        bool         deleted;   // if true -> delete this entry from playlist

    public:
        PlaylistEntry(const std::string& path_, const std::string* extinf_)
            : extinf(extinf_ ? new std::string(*extinf_) : 0)
            , path(path_)
            , deleted(false)
        {}
        PlaylistEntry(const PlaylistEntry& other)
            : extinf(other.extinf ? new std::string(*other.extinf) : 0)
            , path(other.path)
            , deleted(other.deleted)
        {}
        PlaylistEntry& operator = (const PlaylistEntry& other) {
            if (&other!=this) {
                delete(extinf);
                extinf  = other.extinf ? new std::string(*other.extinf) : 0;
                path    = other.path;
                deleted = other.deleted;
            }
            return *this;
        }
        virtual ~PlaylistEntry() { delete extinf; }

        const std::string *Extinf() const { return extinf; }
        void set_Extinf(const std::string *extinf_) {
            delete(extinf);
            extinf = 0;
            if (extinf_) extinf = new std::string(*extinf_);
        }

        const std::string& Path() const { return path; }
        void set_Path(const std::string& path_) { path = path_; }

        void del() { deleted = true; }
        bool is_del() const { return deleted; }

        void write(std::ostream& out, bool pls_format, int index) const;

        bool hasExtInfo() const { return extinf!=0; }
    };

    typedef std::multimap<double, PlaylistEntry> BestEntries;

    //  -----------------------
    //      class PlayList
    //  -----------------------
    class PlayList {
    private:
        std::list<PlaylistEntry> entries;
        std::string         filename;
        bool                pls_format;

        std::string correctExtension() const { return pls_format ? ".pls" : ".m3u"; }

    public:
        PlayList(const std::string& fname) : filename(fname), pls_format(false) {}
        virtual ~PlayList() {}

        const std::string& Filename() const { return filename; }
        void set_Filename(const std::string& filename_) { filename = filename_; }

        bool hasExtInfo() const  {
            for (std::list<PlaylistEntry>::const_iterator e=entries.begin(); e!=entries.end(); ++e) {
                if (e->hasExtInfo()) return true;
            }
            return false;
        }

        void add(const PlaylistEntry& toAdd) { entries.push_back(toAdd); }
        void addSeparator() { entries.push_back(PlaylistEntry("-----------", 0)); }

        bool is_pls_format() const { return pls_format; }
        void set_pls_format(bool pls) { pls_format = pls; }

        void load(std::ostream& screen_out);
        void save(std::ostream& screen_out) const;

        void getBestMatchingEntries(const PlaylistEntry& entry, BestEntries& best, unsigned count) const;

        void addDupesTo(PlayList& dupes, double minMatchValue, int maxTimeDifference);
        int  deleteDuplicateEntries();

        void correct(const PlayList& valid, PlayList& missing, std::ostream& screen_out, bool delDupes, bool autoSelect, int equalPaths, bool copyInfo, bool keepInfo, double askAbove);

    };



}; // end of namespace rs

#else
#error playlist.h included twice
#endif // PLAYLIST_H

