//  ======================================================================== //
//                                                                           //
//    File      : quickSearch.h                                              //
//    Purpose   :                                                            //
//    Time-stamp: <Wed Nov/13/2002 08:45 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) January 2002 by Ralf Westram                                       //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef QUICKSEARCH_H
#define QUICKSEARCH_H

#ifndef LIST
#include <list>
#endif
#ifndef STRING
#include <string>
#endif

namespace rs {

    typedef std::list<size_t> CharOccurances;

#define FIRST_INDEXED_CHAR 32
#define LAST_INDEXED_CHAR  127
#define INDEXED_CHARS      (LAST_INDEXED_CHAR-FIRST_INDEXED_CHAR+1)

    //  --------------------------------
    //      class QuickSearchString
    //  --------------------------------
    class QuickSearchString {
    public:
        static bool is_indexed(int ch) { return ch >= FIRST_INDEXED_CHAR && ch <= LAST_INDEXED_CHAR; }

    private:
        std::string         s;
        CharOccurances occur[INDEXED_CHARS]; // one for each character

        void createIndex() {
            int pos = 0;
            for (std::string::const_iterator c = s.begin(); c != s.end(); ++c, ++pos) {
                char ch = *c;
                if (is_indexed(ch)) occur[ch-FIRST_INDEXED_CHAR].push_back(pos);
            }
        }

    public:
        QuickSearchString(const std::string& s_) : s(s_) { createIndex(); }
        virtual ~QuickSearchString() {}


        CharOccurances::const_iterator begin(unsigned char ch) const {
            assert(is_indexed(ch));
            return occur[ch-FIRST_INDEXED_CHAR].begin();
        }
        CharOccurances::const_iterator end(unsigned char ch) const {
            assert(is_indexed(ch));
            return occur[ch-FIRST_INDEXED_CHAR].end();
        }

        const std::string& getString() const { return s; }
    };

    inline bool isNonWhite(unsigned char ch) { return isalnum(ch) || ch > LAST_INDEXED_CHAR; }

    inline size_t countNonWhite(const std::string& s)  {
        size_t nonWhite = 0;
        for (std::string::const_iterator c = s.begin(); c != s.end(); ++c) {
            nonWhite += isNonWhite(*c);
        }
        return nonWhite;
    }

}; // end of namespace rs

#else
#error quickSearch.h included twice
#endif // QUICKSEARCH_H

